# Third-party
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import numpy as np

# Project
from ..transforms import a_P_to_m, a_m_to_P, P_m_to_a


def test_roundtrip_Pma():

    rnd = np.random.RandomState(seed=42)

    for i in range(128):
        P = 10 ** rnd.uniform(0, 3) * u.day
        m = np.abs(rnd.normal(1., 0.5)) * u.Msun
        a = P_m_to_a(P, m)

        m2 = a_P_to_m(a, P)
        P2 = a_m_to_P(a, m2)
        a2 = P_m_to_a(P2, m2)

        assert quantity_allclose(a, a2)
        assert quantity_allclose(m, m2)
        assert quantity_allclose(P, P2)



# Note: copied out of adrn/gala

# Third-party
import astropy.units as u
from astropy.units.physical import _physical_unit_mapping
import astropy.constants as const

__all__ = ['UnitSystem']


class UnitSystem:
    """
    Represents a system of units. At minimum, this consists of a set of
    length, time, mass, and angle units, but may also contain preferred
    representations for composite units. For example, the base unit system
    could be ``{kpc, Myr, Msun, radian}``, but you can also specify a preferred
    speed, such as ``km/s``.

    This class functions like a dictionary with keys set by physical types.
    If a unit for a particular physical type is not specified on creation,
    a composite unit will be created with the base units. See Examples below
    for some demonstrations.

    Parameters
    ----------
    *units
        The units that define the unit system. At minimum, this must
        contain length, time, mass, and angle units.

    Examples
    --------
    If only base units are specified, any physical type specified as a key
    to this object will be composed out of the base units::

        >>> usys = UnitSystem(u.m, u.s, u.kg, u.radian)
        >>> usys['energy']
        Unit("kg m2 / s2")

    However, custom representations for composite units can also be specified
    when initializing::

        >>> usys = UnitSystem(u.m, u.s, u.kg, u.radian, u.erg)
        >>> usys['energy']
        Unit("erg")

    This is useful for Galactic dynamics where lengths and times are usually
    given in terms of ``kpc`` and ``Myr``, but speeds are given in ``km/s``::

        >>> usys = UnitSystem(u.kpc, u.Myr, u.Msun, u.radian, u.km/u.s)
        >>> usys['speed']
        Unit("km / s")
    """

    def __init__(self, units, *args):

        self._required_physical_types = ['length', 'time', 'mass', 'angle']
        self._core_units = []

        if isinstance(units, UnitSystem):
            self._registry = units._registry.copy()
            self._core_units = units._core_units
            return

        if len(args) > 0:
            units = (units,) + tuple(args)

        self._registry = dict()
        for unit in units:
            typ = unit.physical_type
            if typ in self._registry:
                raise ValueError("Multiple units passed in with type "
                                 "'{0}'".format(typ))
            self._registry[typ] = unit

        for phys_type in self._required_physical_types:
            if phys_type not in self._registry:
                raise ValueError("You must specify a unit with physical type "
                                 "'{0}'".format(phys_type))
            self._core_units.append(self._registry[phys_type])

    def __getitem__(self, key):

        if key in self._registry:
            return self._registry[key]

        else:
            unit = None
            for k,v in _physical_unit_mapping.items():
                if v == key:
                    unit = u.Unit(" ".join(["{}**{}".format(x,y) for x,y in k]))
                    break

            if unit is None:
                raise ValueError("Physical type '{0}' doesn't exist in unit "
                                 "registry.".format(key))

            unit = unit.decompose(self._core_units)
            unit._scale = 1.
            return unit

    def __len__(self):
        return len(self._core_units)

    def __iter__(self):
        for uu in self._core_units:
            yield uu

    def __str__(self):
        return "UnitSystem ({0})".format(
            ",".join([str(uu) for uu in self._core_units]))

    def __repr__(self):
        return "<{0}>".format(self.__str__())

    def __eq__(self, other):
        for k in self._registry:
            if not self[k] == other[k]:
                return False

        for k in other._registry:
            if not self[k] == other[k]:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self):
        """
        Return a dictionary representation of the unit system with keys
        set by the physical types and values set by the unit objects.
        """
        return self._registry.copy()

    def decompose(self, q):
        """
        A thin wrapper around :meth:`astropy.units.Quantity.decompose` that
        knows how to handle Quantities with physical types with non-default
        representations.

        Parameters
        ----------
        q : :class:`~astropy.units.Quantity`
            An instance of an astropy Quantity object.

        Returns
        -------
        q : :class:`~astropy.units.Quantity`
            A new quantity, decomposed to represented in this unit system.
        """
        try:
            ptype = q.unit.physical_type
        except AttributeError:
            raise TypeError("Object must be an astropy.units.Quantity, not "
                            "a '{}'.".format(q.__class__.__name__))

        if ptype in self._registry:
            return q.to(self._registry[ptype])
        else:
            return q.decompose(self)

    def get_constant(self, name):
        """
        Retrieve a constant with specified name in this unit system.

        Parameters
        ----------
        name : str
            The name of the constant, e.g., G.

        Returns
        -------
        const : float
            The value of the constant represented in this unit system.

        Examples
        --------
            >>> usys = UnitSystem(u.kpc, u.Myr, u.radian, u.Msun)
            >>> usys.get_constant('c')
            306.6013937879527
        """
        try:
            c = getattr(const, name)
        except AttributeError:
            raise ValueError("Constant name '{0}' doesn't exist in "
                             "astropy.constants".format(name))

        return c.decompose(self._core_units).value



# coding: utf-8
# cython: boundscheck=False
# cython: nonecheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: profile=False

# Third-party
import numpy as np
cimport numpy as np
np.import_array()
import cython
cimport cython

# from libc.stdio cimport printf

cdef extern from "src/twobody.h":
    double c_mean_anomaly_from_eccentric_anomaly(double E, double e)
    double c_eccentric_anomaly_from_mean_anomaly_Newton1(double M, double e,
                                                         double tol,
                                                         int maxiter)
    double c_eccentric_anomaly_from_mean_anomaly_Householder3(double M,
                                                              double e,
                                                              double tol,
                                                              int maxiter)

    double c_true_anomaly_from_eccentric_anomaly(double E, double e)
    double c_eccentric_anomaly_from_true_anomaly(double f, double e)

    void c_rv_from_elements(double *t, double *rv, int N_t,
                            double P, double K, double e, double omega,
                            double M0, double t0, double tol, int maxiter)

__all__ = ['cy_mean_anomaly_from_eccentric_anomaly',
           'cy_eccentric_anomaly_from_mean_anomaly_Newton1',
           'cy_eccentric_anomaly_from_mean_anomaly_Householder3',
           'cy_true_anomaly_from_eccentric_anomaly',
           'cy_eccentric_anomaly_from_true_anomaly',
           'cy_rv_from_elements']

cpdef cy_mean_anomaly_from_eccentric_anomaly(
        np.ndarray[double, mode="c", ndim=1] E,
        np.ndarray[double, mode="c", ndim=1] e):
    """

    """
    cdef:
        int n
        int N = len(E)
        np.ndarray[double, mode="c", ndim=1] M = np.zeros(N, dtype=np.float64)

    for n in range(N):
        M[n] = c_mean_anomaly_from_eccentric_anomaly(E[n], e[n])

    return M

cpdef cy_eccentric_anomaly_from_mean_anomaly_Newton1(
        np.ndarray[double, mode="c", ndim=1] M,
        np.ndarray[double, mode="c", ndim=1] e,
        double tol, int maxiter):
    """

    """
    cdef:
        int n
        int N = len(M)
        np.ndarray[double, mode="c", ndim=1] E = np.zeros(N, dtype=np.float64)

    for n in range(N):
        E[n] = c_eccentric_anomaly_from_mean_anomaly_Newton1(M[n], e[n],
                                                             tol, maxiter)

    return E

cpdef cy_eccentric_anomaly_from_mean_anomaly_Householder3(
        np.ndarray[double, mode="c", ndim=1] M,
        np.ndarray[double, mode="c", ndim=1] e,
        double tol, int maxiter):
    """

    """
    cdef:
        int n
        int N = len(M)
        np.ndarray[double, mode="c", ndim=1] E = np.zeros(N, dtype=np.float64)

    for n in range(N):
        E[n] = c_eccentric_anomaly_from_mean_anomaly_Householder3(M[n], e[n],
                                                                  tol, maxiter)

    return E

# ----------------------------------------------------------------------------0

cpdef cy_true_anomaly_from_eccentric_anomaly(
        np.ndarray[double, mode="c", ndim=1] E,
        np.ndarray[double, mode="c", ndim=1] e):
    """

    """
    cdef:
        int n
        int N = len(E)
        np.ndarray[double, mode="c", ndim=1] f = np.zeros(N, dtype=np.float64)

    for n in range(N):
        f[n] = c_true_anomaly_from_eccentric_anomaly(E[n], e[n])

    return f

cpdef cy_eccentric_anomaly_from_true_anomaly(
        np.ndarray[double, mode="c", ndim=1] f,
        np.ndarray[double, mode="c", ndim=1] e):
    """

    """
    cdef:
        int n
        int N = len(f)
        np.ndarray[double, mode="c", ndim=1] E = np.zeros(N, dtype=np.float64)

    for n in range(N):
        E[n] = c_eccentric_anomaly_from_true_anomaly(f[n], e[n])

    return E

cpdef cy_rv_from_elements(np.ndarray[double, mode="c", ndim=1] times,
        double P, double K, double e, double omega, double phi0, double t0,
        double anomaly_tol, int anomaly_maxiter):

    cdef:
        int n
        int N_t = len(times)
        double[::1] rv = np.zeros(N_t)

    c_rv_from_elements(&times[0], &rv[0], N_t,
                       P, K, e, omega, phi0, t0,
                       anomaly_tol, anomaly_maxiter)

    return np.array(rv)



# Third-party
import astropy.units as u
from astropy.time import Time
import numpy as np
import pytest

from ..bary_trends import PolynomialRVTrend

def test_polynomial():
    # initialization
    trend = PolynomialRVTrend([10.*u.km/u.s,
                               1.*u.km/u.s/u.day,
                               1.*u.km/u.s/u.day**2])

    trend = PolynomialRVTrend()
    assert trend(0.) == 0.
    assert trend([0.]) == np.array([0.])
    assert np.all(trend(np.arange(10)) == 0.)

    # invalid units
    with pytest.raises(ValueError):
        PolynomialRVTrend([10., 1.*u.km/u.s/u.day])

    with pytest.raises(u.UnitsError):
        PolynomialRVTrend([10.*u.km/u.s, 1.*u.km/u.s])

    trend = PolynomialRVTrend([10.*u.km/u.s, 1.*u.km/u.s/u.day])
    t = Time(55555. + np.sort(np.random.uniform(15., 23., 128)),
             format='mjd')
    res1 = trend(t)
    res2 = trend(t.tcb.mjd)
    assert np.allclose(res1.value, res2.value)
    assert res1.unit == u.km/u.s
    assert res2.unit == u.km/u.s

    trend = PolynomialRVTrend([10.*u.km/u.s, 1.*u.km/u.s/u.day],
                              t0=Time(55555., format='mjd'))
    res = trend(t.tcb.mjd)

    t = (Time(55555., format='mjd') +
         np.sort(np.random.uniform(15., 23., 128))*u.day)
    res = trend(t)



# Third-party
import astropy.coordinates as coord
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import numpy as np

# Project
from ..elements import KeplerElements, TwoBodyKeplerElements
from ..orbit import KeplerOrbit

# TwoBody API exploration
# Note: Anything labeled "Future: v2.0" will be implemented later.

# The Elements classes are used to specify orbital elements. Right now, only the
# KeplerElements class exists and is used to specify Keplerian orbital elements.
elem = KeplerElements(a=1.5*u.au, e=0.5, P=1.*u.year,
                      omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                      M0=53*u.deg, t0=59112.1423)

# The elements correspond to a specific orbit, so if this is one component of a
# binary, a actually corresponds to a1.

# In the above, the epoch, t0, is assumed to be MJD if a number is passed in,
# but it can also be specified as an astropy.time.Time object:
import astropy.time as at
t0 = at.Time(2459812.641, format='jd')
elem = KeplerElements(a=1.5*u.au, e=0.5, P=1.*u.year,
                      omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                      M0=53*u.deg, t0=t0)

# If not specified, the epoch is taken to be J2000:
elem = KeplerElements(a=1.5*u.au, e=0.5, P=1.*u.year,
                      omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg, M0=53*u.deg)
assert elem.t0 == at.Time('J2000')

# These classes support transforming the input elements to other common
# parameters within their element system. For example, with any input (see
# above), the following attributes are available for a KeplerElements instance:
elem.P
elem.a
elem.m_f  # "binary mass function"
elem.K  # velocity semi-amplitude

# To specify orbital elements of a binary or two-body system, use the
# TwoBodyKeplerElements class. With this class, the orbital elements specified
# correspond to the orbit of one of the two bodies in the system. By convention,
# this is usually the more massive "primary" body
# TODO: or should they be the orbital elements of the fictitious particle?
elem = TwoBodyKeplerElements(m1=1*u.Msun, m2=2*u.Msun,
                             a=1.5*u.au, e=0.5,
                             omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                             M0=53*u.deg, t0=59112.1423)

# From this object, we can access the orbital elements of each component orbit
# individually:
elem1 = elem.get_body('1')
elem2 = elem.get_body('2')
# Or, using short-hand names:
# assert elem.primary == elem.get_component('1')
# assert elem.secondary == elem.get_component('2')

assert elem1.__class__ == KeplerElements
assert quantity_allclose(elem1.a,
                         elem.a * elem.m2 / (elem.m1 + elem.m2))
assert quantity_allclose(elem1.omega, elem.omega)
assert quantity_allclose(elem2.omega, elem.omega + np.pi*u.rad)

# We could then create orbit objects for each component in the system:
orb1 = KeplerOrbit(elem.primary)
orb2 = KeplerOrbit(elem.secondary)

# Can also specify P instead of a
elem = TwoBodyKeplerElements(m1=1*u.Msun, m2=2*u.Msun,
                             P=1.*u.year, e=0.5,
                             omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                             M0=53*u.deg, t0=59112.1423)

# Future: v2.0
# - Also support other element parametrizations, e.g., Delaunay variables
#   elem = DelaunayElements(...)

# In the above examples, since we created an Elements instance explicitly, to
# create an Orbit object, we pass the Elements instance in to the KeplerOrbit
# initializer. For example, for a single orbit:
t0 = at.Time(2459812.641, format='jd')
elem = KeplerElements(a=1.5*u.au, e=0.5, P=1.*u.year,
                      omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                      M0=53*u.deg, t0=t0)
orb = KeplerOrbit(elements=elem)
orb = KeplerOrbit(elem)

# We don't have to create an Elements object before creating an orbit. Using the
# string name of the elements class, we can pass the components directly in to
# the KeplerOrbit initializer:
orb = KeplerOrbit(a=1.5*u.au, e=0.5, P=1.*u.year,
                  omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                  M0=53*u.deg, t0=t0,
                  elements_type='kepler')  # this is default

# Because 'kepler' is default, this will work:
orb = KeplerOrbit(a=1.5*u.au, e=0.5, P=1.*u.year,
                  omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                  M0=53*u.deg, t0=t0)

# With an orbit instance, we can easily access observable orbital quantities,
# for example, to get the radial velocity of the body (corresponding to `m1`)
time = at.Time([59812.141, 59142.782, 59147.559], format='mjd')
rv = orb.radial_velocity(time)

# If instead we didn't specify any masses, the computed orbital properties [are
# dimensionless TODO?] but can be scaled to a velocity. For example, the radial
# velocity curve can be directly multiplied by a semi-amplitude:
orb = KeplerOrbit(P=1.*u.year, e=0.5,
                  omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
                  M0=53*u.deg, t0=t0)
K = 10 * u.m/u.s
rv = K * orb.unscaled_radial_velocity(time)



# Standard library
import inspect

# Third-party
import astropy.units as u
from astropy.utils.decorators import wraps

# Project
from .wrap import (cy_mean_anomaly_from_eccentric_anomaly,
                   cy_true_anomaly_from_eccentric_anomaly,
                   cy_eccentric_anomaly_from_true_anomaly,
                   cy_eccentric_anomaly_from_mean_anomaly_Newton1,
                   cy_eccentric_anomaly_from_mean_anomaly_Householder3)
from .utils import ArrayProcessor

__all__ = ['mean_anomaly_from_eccentric_anomaly',
           'eccentric_anomaly_from_mean_anomaly',
           'true_anomaly_from_eccentric_anomaly',
           'eccentric_anomaly_from_true_anomaly']


def anomaly_wrapper(func):
    """This is a decorator that processes the input (strips units, enforces that
    it is a 1D array), executes the cython function on the cleaned input, then
    returns the output angle in the same units and with the same shape as the
    input.
    """
    sig = inspect.signature(func)

    @wraps(func)
    def function_wrapper(ang, e, *args, **kwargs):
        in_unit = ang.unit
        p = ArrayProcessor(ang.to(u.radian).value, e)
        ang, e = p.prepare_arrays()
        res = p.prepare_result(func(ang, e, *args, **kwargs))
        return (res * u.radian).to(in_unit)

    return function_wrapper


@u.quantity_input(E=u.radian)
@anomaly_wrapper
def mean_anomaly_from_eccentric_anomaly(E, e):
    """
    Parameters
    ----------
    E : quantity_like [angle]
        Eccentric anomaly.
    e : numeric, array_like
        Eccentricity.

    Returns
    -------
    M : numeric, array_like [radian]
        Mean anomaly.
    """
    return cy_mean_anomaly_from_eccentric_anomaly(E, e)


@u.quantity_input(M=u.radian)
@anomaly_wrapper
def eccentric_anomaly_from_mean_anomaly(M, e, tol=1E-10, maxiter=128,
                                        method='Newton1'):
    """
    Parameters
    ----------
    M : quantity_like [angle]
        Mean anomaly.
    e : numeric
        Eccentricity.
    tol : numeric, optional
        Numerical tolerance used in iteratively solving for eccentric anomaly.
    maxiter : int, optional
        Maximum number of iterations when iteratively solving for eccentric
        anomaly.
    method : str, optional
        The method to use for iterative root-finding for the eccentric anomaly.
        Options are: ``'Newton1'`` and ``'Householder3'``.

    Returns
    -------
    E : numeric [radian]
        Eccentric anomaly.

    Issues
    ------
    - Magic numbers ``tol`` and ``maxiter``
    """
    func = eval('cy_eccentric_anomaly_from_mean_anomaly_{0}'.format(method))
    return func(M, e, tol, maxiter)


@u.quantity_input(E=u.radian)
@anomaly_wrapper
def true_anomaly_from_eccentric_anomaly(E, e):
    """
    Parameters
    ----------
    E : quantity_like [angle]
        Eccentric anomaly.
    e : numeric, array_like
        Eccentricity.

    Returns
    -------
    f : numeric [radian]
        True anomaly.
    """
    return cy_true_anomaly_from_eccentric_anomaly(E, e)


@u.quantity_input(f=u.radian)
@anomaly_wrapper
def eccentric_anomaly_from_true_anomaly(f, e):
    """
    Parameters
    ----------
    f : quantity_like [angle]
        True anomaly.
    e : numeric, array_like
        Eccentricity.

    Returns
    -------
    E : numeric [radian]
        Eccentric anomaly.
    """
    return cy_eccentric_anomaly_from_true_anomaly(f, e)



# Third-party
from astropy.tests.helper import quantity_allclose
import astropy.units as u
import numpy as np
import pytest

# Project
from ..anomaly import (mean_anomaly_from_eccentric_anomaly,
                       eccentric_anomaly_from_mean_anomaly,
                       true_anomaly_from_eccentric_anomaly,
                       eccentric_anomaly_from_true_anomaly)

N = 16


def test_mean_anomaly_from_eccentric_anomaly():
    E_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1, N)
    Es, es = [x.ravel() for x in np.meshgrid(E_grid, e_grid)]

    # call with float
    for E, e in zip(Es, es):
        M = mean_anomaly_from_eccentric_anomaly(E*u.rad, e)
        assert M.shape == ()

    # call with arrays
    M = mean_anomaly_from_eccentric_anomaly(Es*u.rad, es)
    assert M.shape == Es.shape


@pytest.mark.parametrize("method", ["Newton1", "Householder3"])
def test_eccentric_anomaly_from_mean_anomaly(method):
    M_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1, N)
    Ms, es = [x.ravel() for x in np.meshgrid(M_grid, e_grid)]

    # call with float
    for M, e in zip(Ms, es):
        E = eccentric_anomaly_from_mean_anomaly(M*u.rad, e, method=method)
        assert E.shape == ()

    # call with arrays
    E = eccentric_anomaly_from_mean_anomaly(Ms*u.rad, es)
    assert E.shape == Ms.shape


@pytest.mark.parametrize("method", ["Newton1", "Householder3"])
def test_anomaly_roundtrip(method):
    kw = dict(tol=1E-10, maxiter=128, method=method)

    # M -> E -> M
    M_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1 - 1E-3, N)  # MAGIC NUMBER: 1E-3
    Ms, es = [x.ravel() for x in np.meshgrid(M_grid, e_grid)]

    for M, e in zip(Ms, es):
        E = eccentric_anomaly_from_mean_anomaly(M*u.rad, e, **kw)
        M2 = mean_anomaly_from_eccentric_anomaly(E, e)
        assert quantity_allclose(M*u.rad, M2, atol=1E-14*u.rad)

    E = eccentric_anomaly_from_mean_anomaly(Ms*u.rad, es, **kw)
    M2 = mean_anomaly_from_eccentric_anomaly(E, es)
    assert quantity_allclose(Ms*u.rad, M2, atol=1E-14*u.rad)

    # E -> M -> E
    E_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1 - 1E-3, N)  # MAGIC NUMBER: 1E-3
    Es, es = [x.ravel() for x in np.meshgrid(E_grid, e_grid)]

    for E, e in zip(Es, es):
        M = mean_anomaly_from_eccentric_anomaly(E*u.rad, e)
        E2 = eccentric_anomaly_from_mean_anomaly(M, e, **kw)
        assert quantity_allclose(E*u.rad, E2, atol=1E-14*u.rad)

    M = mean_anomaly_from_eccentric_anomaly(Es*u.rad, es)
    E2 = eccentric_anomaly_from_mean_anomaly(M, es, **kw)
    assert quantity_allclose(Es*u.rad, E2, atol=1E-14*u.rad)


def test_true_anomaly_roundtrip():
    # f -> E -> f
    f_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1 - 1E-3, N)  # MAGIC NUMBER: 1E-3
    fs, es = [x.ravel() for x in np.meshgrid(f_grid, e_grid)]

    E = eccentric_anomaly_from_true_anomaly(fs*u.rad, es)
    f2 = true_anomaly_from_eccentric_anomaly(E, es)
    assert np.allclose(np.cos(fs), np.cos(f2), atol=1E-9)

    # E -> f -> E
    E_grid = np.linspace(0, 2*np.pi, N)
    e_grid = np.linspace(0, 1 - 1E-3, N)  # MAGIC NUMBER: 1E-3
    Es, es = [x.ravel() for x in np.meshgrid(E_grid, e_grid)]

    f = true_anomaly_from_eccentric_anomaly(Es*u.rad, es)
    E2 = eccentric_anomaly_from_true_anomaly(f, es)
    assert np.allclose(np.cos(Es), np.cos(E2), atol=1E-9)



# Third-party
from astropy.time import Time
import astropy.units as u
import astropy.coordinates as coord
from astropy.tests.helper import catch_warnings
import pytest

# Project
from ..barycenter import Barycenter


def test_init():

    b = Barycenter(origin=coord.SkyCoord(ra=150 * u.deg,
                                         dec=-11 * u.deg,
                                         distance=100 * u.pc))

    b = Barycenter(origin=coord.SkyCoord(ra=150 * u.deg,
                                         dec=-11 * u.deg,
                                         distance=100 * u.pc),
                   t0=59123.1213)

    b = Barycenter(origin=coord.SkyCoord(ra=150 * u.deg,
                                         dec=-11 * u.deg,
                                         distance=100 * u.pc),
                   t0=Time(59123.1213, format='mjd'))

    c = coord.SkyCoord(ra=150 * u.deg, dec=-11 * u.deg, distance=100 * u.pc)
    with pytest.raises(ValueError):
        Barycenter(origin=c, radial_velocity=100*u.km/u.s)

    with pytest.raises(ValueError):
        Barycenter()

    with pytest.raises(TypeError):
        Barycenter(origin=Time('J2000'))

    with pytest.raises(ValueError):
        Barycenter(origin=coord.SkyCoord(ra=[150, 100] * u.deg,
                                         dec=[-11, 10] * u.deg,
                                         distance=100 * u.pc))

    with catch_warnings() as w:
        b = Barycenter(origin=coord.SkyCoord(ra=150 * u.deg,
                                             dec=-11 * u.deg))
    assert len(w) > 0



# Standard library
import warnings

# Third-party
import astropy.coordinates as coord
from astropy.time import Time
import astropy.units as u
import numpy as np

# Project
from .elements import _parse_time

__all__ = ['Barycenter']


class Barycenter:

    def __init__(self, origin=None, t0=None, radial_velocity=None):
        """The location of the barycenter at the specified epoch.

        This class is used to specify a reference location and velocity of the
        barycenter at a given epoch. This is used for computing astrometric
        orbits, and for specifying the barycentric motion.

        In the future, this class will contain more functionality that supports,
        e.g., acceleration of the barycenter from a third body.

        Parameters
        ----------
        origin : `~astropy.coordinates.SkyCoord` or frame instance
            The sky position of and distance to the barycenter at the specified
            epoch, ``t0``. If only a barycentric radial velocity is known or
            needed, pass in ``radial_velocity``.
        t0 : numeric, `~astropy.coordinates.Time` (optional)
            Reference epoch for the location of the barycenter. If a number is
            passed in, it is assumed to be a solar system barycentric modified
            julian date (BMJD). The default is J2000 if not specified.
        radial_velocity : quantity_like (optional)
            If the coordinates or distance to the barycenter are not known or
            needed, you can just pass the radial velocity of the barycenter
            instead.

        Examples
        --------
        >>> import astropy.units as u
        >>> from astropy.coordinates import SkyCoord
        >>> b = Barycenter(radial_velocity=100*u.km/u.s)
        >>> b = Barycenter(origin=SkyCoord(ra=150*u.deg, dec=37*u.deg,
        ...                                distance=100*u.pc)) # assumes J2000
        >>> from astropy.time import Time
        >>> b = Barycenter(origin=SkyCoord(ra=150*u.deg, dec=37*u.deg,
        ...                                distance=100*u.pc),
        ...                t0=Time(58912.481293, format='mjd'))

        """

        if origin is not None and radial_velocity is not None:
            raise ValueError("Pass one of: origin or radial_velocity, not both")

        if origin is None and radial_velocity is None:
            raise ValueError("You must pass one of origin or radial_velocity")

        if origin is None:
            origin = coord.ICRS(ra=0*u.deg, dec=0*u.deg, distance=np.nan*u.pc,
                                radial_velocity=radial_velocity)

        if (not isinstance(origin, coord.SkyCoord) and
                not isinstance(origin, coord.BaseCoordinateFrame)):
            raise TypeError("origin must be a SkyCoord or BaseCoordinateFrame "
                            "subclass instance, i.e. a valid "
                            "astropy.coordinates object.")

        if not origin.isscalar:
            raise ValueError("origin must be a scalar coordinate, not "
                             "array-valued: {0}".format(origin))

        if origin.distance.unit == u.dimensionless_unscaled:
            warnings.warn("Barycenter origin has no associated distance. Are "
                          "you sure you know what you're doing?",
                          UserWarning)

        self.origin = origin

        if t0 is None:
            t0 = Time('J2000')
        self.t0 = _parse_time(t0)

    def __repr__(self):
        fmt_str = "<Barycenter: origin={0.origin}, epoch={0.t0}>".format(self)
        return fmt_str

    def __str__(self):
        return repr(self)



# Third-party
import astropy.units as u
import astropy.coordinates as coord
from astropy.tests.helper import quantity_allclose
import numpy as np
import pytest

# Project
from ..reference_plane import ReferencePlaneFrame


def test_sanity():

    rep1 = coord.CartesianRepresentation(x=[0, 1.],
                                         y=0,
                                         z=0, unit=u.pc)

    rep2 = coord.CartesianRepresentation(x=0,
                                         y=[0, 1.],
                                         z=0, unit=u.pc)

    rep3 = coord.CartesianRepresentation(x=0,
                                         y=0,
                                         z=[0, 1.], unit=u.pc)

    # Try for many origins:
    rnd = np.random.RandomState(seed=42)

    for _ in range(128):
        origin = coord.ICRS(ra=rnd.uniform(0, 360)*u.deg,
                            dec=rnd.uniform(-90, 90)*u.deg,
                            distance=rnd.uniform(10, 100)*u.pc)

        ref_c = ReferencePlaneFrame(rep1, origin=origin)
        icrs = ref_c.transform_to(coord.ICRS)
        assert icrs.dec[1] > icrs.dec[0]
        assert quantity_allclose(icrs.ra[0], icrs.ra[1])

        ref_c = ReferencePlaneFrame(rep2, origin=origin)
        icrs = ref_c.transform_to(coord.ICRS)
        assert icrs.ra[1] > icrs.ra[0]

        ref_c = ReferencePlaneFrame(rep3, origin=origin)
        icrs = ref_c.transform_to(coord.ICRS)
        assert icrs.distance[0] > icrs.distance[1]
        assert quantity_allclose(icrs.ra[0], icrs.ra[1])
        assert quantity_allclose(icrs.dec[0], icrs.dec[1])


def test_transform():
    rep = coord.CartesianRepresentation(x=1, y=2, z=3, unit=u.pc)
    ref_c1 = ReferencePlaneFrame(rep)

    with pytest.raises(ValueError):
        ref_c1.transform_to(coord.ICRS)

    with pytest.raises(ValueError):
        ref_c2 = ReferencePlaneFrame(rep, origin=coord.Galactic())



# Third-party
import astropy.coordinates as coord
from astropy.coordinates.matrix_utilities import matrix_product, rotation_matrix
from astropy.time import Time
import astropy.units as u
import numpy as np
from numpy import pi

# Project
from . import elements as elem
from .anomaly import (eccentric_anomaly_from_mean_anomaly,
                      true_anomaly_from_eccentric_anomaly)
from .utils import ArrayProcessor
from .barycenter import Barycenter
from .wrap import cy_rv_from_elements
from .reference_plane import ReferencePlaneFrame
from .bary_trends import RVTrend, PolynomialRVTrend

__all__ = ['KeplerOrbit']

_KMS = u.km/u.s

class KeplerOrbit:

    def __init__(self, elements=None, elements_type='kepler',
                 barycenter=None, **kwargs):
        """Represents a bound Kepler orbit.

        Parameters
        ----------
        elements : `twobody.OrbitalElements` subclass instance
            Either pass in an ``OrbitalElements`` object, e.g., an instance of
            `twobody.KeplerElements`, or pass in the element names themselves.
            If the latter, anything passed in as kwargs gets passed to the
            elements class specified by ``elements_type``. The element names
            for the default ``elements_type`` are included below for
            convenience.
        elements_type : str (optional)
            Ignore if you pass in an instantiated ``OrbitalElements`` object.
            This argument controls the class that the ``kwargs`` are passed to.
            The default is ``'kepler'``, meaning all keyword arguments get
            passed to the `twobody.KeplerElements` class.
        barycenter : `twobody.Barycenter` (optional)
            Parameters that control specification of the barycenter of the
            orbit.

        Kepler elements
        ---------------
        P : quantity_like [time]
            Orbital period.
        a : quantity_like [length] (optional)
            Semi-major axis. Specify this OR the semi-amplitude ``K``, but not
            both. If unspecified, computed orbits will be unscaled.
        K : quantity_like [speed] (optional)
            Velocity semi-amplitudes. Specify this OR the semi-major axis ``a``,
            but not both. If unspecified, computed orbits will be unscaled.
        e : numeric (optional)
            Orbital eccentricity. Default is circular, ``e=0``.
        omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Argument of pericenter.
        i : quantity_like, `~astropy.coordinates.Angle` [angle]
            Inclination of the orbit.
        Omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Longitude of the ascending node.
        M0 : quantity_like, `~astropy.coordinates.Angle` [angle] (optional)
            Mean anomaly at epoch ``t0``. Default is 0º if not specified.
        t0 : numeric, `~astropy.coordinates.Time` (optional)
            Reference epoch. If a number is passed in, it is assumed to be
            a solar system barycentric modified julian date (BMJD). The default
            is J2000 if not specified.
        units : `~twobody.units.UnitSystem`, iterable (optional)
            The unit system to represent quantities in. The default unit system
            is accessible as `KeplerElements.default_units`.

        Examples
        --------
        As described above, you can either create an ``Elements`` object and
        then pass this to ``KeplerOrbit``, e.g.,

            >>> import astropy.units as u
            >>> from astropy.time import Time
            >>> from twobody import KeplerElements
            >>> t0 = Time(2459812.641, format='jd') # reference epoch
            >>> elem = KeplerElements(a=1.5*u.au, e=0.5, P=1.*u.year,
            ...                       omega=67*u.deg, i=21.*u.deg,
            ...                       Omega=33*u.deg, M0=53*u.deg, t0=t0)
            >>> orb = KeplerOrbit(elem)

        Or, you can pass in the element names as arguments to the
        ``KeplerOrbit`` class:

            >>> orb = KeplerOrbit(a=1.5*u.au, e=0.5, P=1.*u.year,
            ...                   omega=67*u.deg, i=21.*u.deg, Omega=33*u.deg,
            ...                   M0=53*u.deg, t0=t0)
        """

        if elements is None:
            elements_cls = getattr(elem,
                                   "{0}Elements"
                                   .format(elements_type.capitalize()))

            # pass everything in kwargs to the class initializer
            elements = elements_cls(**kwargs)

        elif not isinstance(elements, elem.OrbitalElements):
            raise TypeError("'elements' must be an instance of an "
                            "OrbitalElements subclass.")

        if barycenter is not None and not isinstance(barycenter, (RVTrend,
                                                                  Barycenter)):
            raise TypeError("barycenter must be a twobody.Barycenter instance")

        self.elements = elements

        if isinstance(barycenter, RVTrend):
            self._vtrend = barycenter
            barycenter = None

        elif barycenter is None:
            self._vtrend = lambda t: 0

        else:
            self._vtrend = PolynomialRVTrend([
                barycenter.origin.radial_velocity])

        self._barycenter = barycenter

    ##########################################################################
    # Read-only attributes

    @property
    def barycenter(self):
        return self._barycenter

    ##########################################################################
    # Python special methods

    def __getattr__(self, name):

        # This gives access to the orbital element components directly from the
        # Orbit instance
        if hasattr(self.elements, name):
            return getattr(self.elements, name)

        else:
            raise AttributeError("type object '{0}' has no attribute '{1}'"
                                 .format(self.__class__.__name__, name))

    def __copy__(self):
        import copy
        elements = copy.copy(self.elements)
        barycen = copy.copy(self.barycenter)
        return self.__class__(elements=elements, barycenter=barycen)

    def unscaled_radial_velocity(self, time,
                                 anomaly_tol=None, anomaly_maxiter=None):
        """Compute the unscaled radial velocity of the body at the specified
        times relative to the barycenter or reference point, i.e. in the
        reference plane system not in a solar system barycentric frame.

        See the docstring of `~twobody.KeplerOrbit.radial_velocity` for more
        information and caveats.

        Parameters
        ----------
        time : array_like, `astropy.time.Time`
            Array of times as barycentric MJD values, or an Astropy
            `~astropy.time.Time` object containing the times to evaluate at.
        anomaly_tol : numeric (optional)
            Tolerance passed to
            `~twobody.eccentric_anomaly_from_mean_anomaly` for solving
            for the eccentric anomaly. See default value in that function.
        anomaly_maxiter : numeric (optional)
            Maximum number of iterations to use in
            `~twobody.eccentric_anomaly_from_mean_anomaly` for solving
            for the eccentric anomaly. See default value in that function.

        Returns
        -------
        rv : numeric [m/s]
            Relative radial velocity - does not include systemtic velocity!
        """
        if anomaly_tol is None:
            # TODO: make this a config item?
            anomaly_tol = 1E-10

        if anomaly_maxiter is None:
            # TODO: make this a config item?
            anomaly_maxiter = 128

        # TODO: do we always want to use MJD? precision issues...
        time = time.tcb.mjd
        proc = ArrayProcessor(time)
        t, = proc.prepare_arrays()
        rv = cy_rv_from_elements(t, self.P.to(u.day).value, 1., self.e,
                                 self.omega.to(u.radian).value,
                                 self.M0.to(u.radian).value,
                                 self.t0.tcb.mjd,
                                 anomaly_tol, anomaly_maxiter)
        return np.atleast_1d(proc.prepare_result(rv))

    def radial_velocity(self, time, anomaly_tol=None, anomaly_maxiter=None):
        """Compute the radial velocity of the body at the specified times
        relative to the barycenter or reference point, i.e. in the reference
        plane system not in a solar system barycentric frame.

        This should always be close (in a machine precision sense) to
        the ``z`` velocity of ``orbit.reference_plane(time).``

        When the barycenter is assumed to be at rest with respect to tangential
        motion relative to the observer, this should be equivalent to
        ``orbit.icrs(time).radial_velocity``

        As mentioned above and in :ref:`getting-started-rv`, the radial velocity
        computed this way assumes that the barycenter does not move tangentially
        between epochs and thus ignores spherical projection effects. For
        sources with large proper motions, the true observable line-of-sight
        velocity will change slightly over time. We can visualize the expected
        differences given a full specification of the position and motion of the
        barycenter:

        >>> import astropy.units as u
        >>> import astropy.coordinates as coord
        >>> from astropy.time import Time
        >>> from twobody import Barycenter, KeplerOrbit
        >>> origin = coord.ICRS(ra=170.8743*u.deg, dec=-71.34*u.deg,
        ...                     distance=57.134*u.pc,
        ...                     pm_ra_cosdec=-206.718*u.mas/u.yr,
        ...                     pm_dec=301.82*u.mas/u.yr,
        ...                     radial_velocity=41.84*u.km/u.s)
        >>> baryc = Barycenter(origin=origin, t0=Time('J2000'))
        >>> orb = KeplerOrbit(P=1.5*u.year, e=0.67, a=1.77*u.au,
        ...                   omega=17.14*u.deg, i=65*u.deg, Omega=0*u.deg,
        ...                   M0=35.824*u.deg, t0=Time('J2015.0'),
        ...                   barycenter=baryc)
        >>> t = Time('J2000') + np.linspace(0, 15, 10000) * u.year
        >>> true_rv = orb.icrs(t).radial_velocity
        >>> approx_rv = orb.radial_velocity(t)
        >>> (true_rv - approx_rv).to(u.m/u.s).max() # doctest: +FLOAT_CMP
        <Quantity 3.315196290913036 m / s>

        In this case, the maximum difference is only ~3 m/s.

        Parameters
        ----------
        time : array_like, `astropy.time.Time`
            Array of times as barycentric MJD values, or an Astropy
            `~astropy.time.Time` object containing the times to evaluate at.
        anomaly_tol : numeric (optional)
            Tolerance passed to
            `~twobody.eccentric_anomaly_from_mean_anomaly` for solving
            for the eccentric anomaly. See default value in that function.
        anomaly_maxiter : numeric (optional)
            Maximum number of iterations to use in
            `~twobody.eccentric_anomaly_from_mean_anomaly` for solving
            for the eccentric anomaly. See default value in that function.
        """
        trend = self._vtrend(time)
        rv = self.K * self.unscaled_radial_velocity(time) + trend
        if not rv.unit.is_equivalent(_KMS):
            raise ValueError('Orbit does not have enough valid orbital '
                             'element information to compute a unit-ful '
                             'radial velocity. Use '
                             '`unscaled_radial_velocity()` manually, or '
                             're-initialize with full information (e.g., '
                             'with all angles specified: i, Omega)')
        return rv

    def orbital_plane(self, time):
        """Compute the orbit at specified times in the two-body barycentric
        frame aligned with the orbital plane (xyz).

        Parameters
        ----------
        time : array_like, `astropy.time.Time`
            Array of times as barycentric MJD values, or an Astropy
            `~astropy.time.Time` object containing the times to evaluate at.
        """

        # mean anomaly
        with u.set_enabled_equivalencies(u.dimensionless_angles()):
            M = 2*pi * (time.tcb - self.t0.tcb) / self.P - self.M0
            M = M.to(u.radian)

        # eccentric anomaly
        E = eccentric_anomaly_from_mean_anomaly(M, self.e)

        # true anomaly
        f = true_anomaly_from_eccentric_anomaly(E, self.e)

        # distance from center of mass to orbiting body
        r = self.a * (1. - self.e * np.cos(E))

        # compute the orbit in the cartesian, orbital plane system (xyz):
        x = r * np.cos(f)
        y = r * np.sin(f)
        z = np.zeros_like(x)

        fac = 2*pi * self.a / self.P / np.sqrt(1 - self.e**2)
        vx = -fac * np.sin(f)
        vy = fac * (np.cos(f) + self.e)
        vz = np.zeros_like(vx)

        xyz = coord.CartesianRepresentation(x=x, y=y, z=z)
        vxyz = coord.CartesianDifferential(d_x=vx, d_y=vy, d_z=vz)

        return xyz.with_differentials(vxyz)

    def reference_plane(self, time):
        """Compute the orbit at specified times in the two-body barycentric
        frame aligned with the reference plane coordinate system (XYZ).

        Parameters
        ----------
        time : array_like, `astropy.time.Time`
            Array of times as barycentric MJD values, or an Astropy
            `~astropy.time.Time` object containing the times to evaluate at.
        """

        xyz = self.orbital_plane(time)
        vxyz = xyz.differentials['s']
        xyz = xyz.without_differentials()

        # Construct rotation matrix to take the orbit from the orbital plane
        # system (xyz) to the reference plane system (XYZ):
        R1 = rotation_matrix(-self.omega, axis='z')
        R2 = rotation_matrix(self.i, axis='x')
        R3 = rotation_matrix(self.Omega, axis='z')
        Rot = matrix_product(R3, R2, R1)

        # Rotate to the reference plane system
        XYZ = coord.CartesianRepresentation(matrix_product(Rot, xyz.xyz))
        VXYZ = coord.CartesianDifferential(matrix_product(Rot, vxyz.d_xyz))
        XYZ = XYZ.with_differentials(VXYZ)

        kw = dict()
        if self.barycenter is not None:
            kw['origin'] = self.barycenter.origin
        return ReferencePlaneFrame(XYZ, **kw)

    def icrs(self, time):
        """Return the ICRS (i.e. reference plane) position and velocity of the
        orbit at the specified time(s).

        Parameters
        ----------
        time : array_like, `~astropy.time.Time`
            Time array. Either in BMJD or as an Astropy time.
        """
        rp = self.reference_plane(time)

        icrs_cart = rp.transform_to(coord.ICRS).cartesian
        icrs_pos = icrs_cart.without_differentials()
        icrs_vel = icrs_cart.differentials['s']

        bary_cart = self.barycenter.origin.cartesian
        bary_vel = bary_cart.differentials['s']

        dt = time - self.barycenter.t0
        dx = (bary_vel * dt).to_cartesian()

        new_pos = icrs_pos + dx
        new_vel = icrs_vel + bary_vel

        return coord.ICRS(new_pos.with_differentials(new_vel))

    def plot_rv(self, time, ax=None, rv_unit=None, t_kwargs=None,
                plot_kwargs=None):
        """Plot the line-of-sight or radial velocity at the specified times.

        Parameters
        ----------
        time : array_like, `~astropy.time.Time`
            Time array. Either in BMJD or as an Astropy time.
        ax : `~matplotlib.axes.Axes`, optional
            The axis to draw on (default is to grab the current
            axes using `~matplotlib.pyplot.gca`).
        rv_unit : `~astropy.units.UnitBase`, optional
            Units to plot the radial velocities in (default is km/s).
        t_kwargs : dict, optional
            Keyword arguments passed to :class:`astropy.time.Time` with the
            input time array. For example, ``dict(format='mjd', scale='tcb')``
            for Barycentric MJD.
        plot_kwargs : dict, optional
            Any additional arguments or style settings passed to
            :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes object that the RV curve was drawn on.

        """

        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if rv_unit is None:
            rv_unit = u.km / u.s

        if t_kwargs is None:
            t_kwargs = dict(format='mjd', scale='tcb')

        if plot_kwargs is None:
            plot_kwargs = dict()

        style = plot_kwargs.copy()
        style.setdefault('linestyle', '-')
        style.setdefault('alpha', 0.5)
        style.setdefault('marker', None)

        if not isinstance(time, Time):
            time = Time(time, **t_kwargs)
        rv = self.radial_velocity(time).to(rv_unit).value

        _t = getattr(getattr(time, t_kwargs['scale']), t_kwargs['format'])
        ax.plot(_t, rv, **style)

        return ax



import astropy.coordinates as coord
import astropy.units as u
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from astropy.time import Time
from twobody import KeplerOrbit, TwoBodyKeplerElements, ReferencePlaneFrame, Barycenter

elem = TwoBodyKeplerElements(P=1.5*u.year, e=0.67, M0=0*u.deg,
                             omega=17.14*u.deg, i=65*u.deg, Omega=0*u.deg,
                             m1=1*u.Msun, m2=2*u.Msun)

orb1 = KeplerOrbit(elem.primary)
orb2 = KeplerOrbit(elem.secondary)

t = Time('J2000') + np.linspace(0, 1, 10000) * elem.P

rp1 = orb1.reference_plane(t)
rp2 = orb2.reference_plane(t)

fig, ax = plt.subplots(1, 1, figsize=(5,5))

ax.plot(rp1.cartesian.x, rp1.cartesian.y,
        color='tab:blue', lw=10, marker='')
ax.plot(rp2.cartesian.x, rp2.cartesian.y,
        color='tab:red', lw=10, marker='')

i = 3100
ax.plot(rp1.cartesian.x[i], rp1.cartesian.y[i],
        marker='o', ms=75, color='tab:blue', mew=3, mec='w')
ax.plot(rp2.cartesian.x[i], rp2.cartesian.y[i],
        marker='o', ms=75, color='tab:red', mew=3, mec='w')

ax.scatter(0, 0, marker='o', s=80, color='k')

circ = mpl.patches.Circle((-0.5,0), 2.1, zorder=-100, color='w')
ax.add_patch(circ)

ax.set_aspect('equal', 'datalim')

ax.xaxis.set_visible(False)
ax.yaxis.set_visible(False)

ax.set_xlim(-2.6, 1.6)
ax.set_ylim(-2.1, 2.1)

for k in ax.spines:
    ax.spines[k].set_visible(False)

ax.set_facecolor('none')
fig.set_facecolor('none')

fig.tight_layout()

fig.savefig('icon.png', dpi=250, facecolor=fig.get_facecolor())
fig.savefig('icon.ico', format='png', dpi=50, facecolor=fig.get_facecolor())



import numpy as np
cimport numpy as np
np.import_array()

cpdef cy_mean_anomaly_from_eccentric_anomaly(
    np.ndarray[double, mode="c", ndim=1] E,
    np.ndarray[double, mode="c", ndim=1] e)

cpdef cy_eccentric_anomaly_from_mean_anomaly_Newton1(
    np.ndarray[double, mode="c", ndim=1] M,
    np.ndarray[double, mode="c", ndim=1] e,
    double tol, int maxiter)

cpdef cy_eccentric_anomaly_from_mean_anomaly_Householder3(
    np.ndarray[double, mode="c", ndim=1] M,
    np.ndarray[double, mode="c", ndim=1] e,
    double tol, int maxiter)

cpdef cy_true_anomaly_from_eccentric_anomaly(
    np.ndarray[double, mode="c", ndim=1] E,
    np.ndarray[double, mode="c", ndim=1] e)

cpdef cy_eccentric_anomaly_from_true_anomaly(
    np.ndarray[double, mode="c", ndim=1] E,
    np.ndarray[double, mode="c", ndim=1] e)



# Third-party
from astropy.time import Time
import astropy.units as u
from astropy.tests.helper import catch_warnings
import numpy as np
import pytest

# Project
from ..elements import KeplerElements, TwoBodyKeplerElements
from ..orbit import KeplerOrbit
from ..units import UnitSystem


def test_kepler():

    # Valid initialization:
    kw1 = dict(P=10*u.day, e=0.5,
               omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)
    kw2 = dict(a=0.1*u.au)
    kw3 = dict(M0=40*u.deg, t0=Time('J2015.0'))
    kw4 = dict(units=UnitSystem(u.m, u.s, u.rad, u.kg))
    kw5 = dict(K=10*u.km/u.s)

    for kw in [kw1, {**kw1, **kw2}, {**kw1, **kw2, **kw3}, {**kw1, **kw3},
               {**kw1, **kw4}, {**kw1, **kw2, **kw4},
               {**kw1, **kw2, **kw3, **kw4},
               {**kw1, **kw5}]:  # not meant to be exhaustive
        elems = KeplerElements(**kw)

        # Make sure this also works as arguments to KeplerOrbit
        KeplerOrbit(elems, elements_type='kepler')
        KeplerOrbit(**kw, elements_type='kepler')

    # Invalid initialization:
    for name in ['P', 'omega']:
        kw = kw1.copy()
        kw.pop(name)
        with pytest.raises(ValueError):
            KeplerElements(**kw)

    # check they get set to nan:
    kw = kw1.copy()
    kw.pop('Omega')
    elem = KeplerElements(**kw)
    assert np.isnan(elem.Omega)

    kw = kw1.copy()
    kw.pop('i')
    elem = KeplerElements(**kw)
    assert np.isnan(elem.i)

    # Make sure attributes exist and have correct units:
    elems = KeplerElements(P=10*u.day, e=0.5, a=0.1*u.au,
                           omega=10*u.deg, i=20*u.deg, Omega=30*u.deg,
                           units=UnitSystem(u.m, u.s, u.rad, u.kg, u.m/u.s))

    assert elems.P.unit == u.s
    assert elems.K.unit == u.m/u.s
    assert elems.m_f.unit == u.kg

    elems = KeplerElements(P=10*u.day, e=0.5, K=10*u.km/u.s,
                           omega=10*u.deg, i=20*u.deg, Omega=30*u.deg,
                           units=UnitSystem(u.m, u.s, u.rad, u.kg, u.m/u.s))

    assert elems.P.unit == u.s
    assert elems.K.unit == u.m/u.s
    assert elems.K == 10*u.km/u.s
    assert elems.m_f.unit == u.kg

    # Expected failures:
    bad_kws = [('P', -10*u.day), ('a', -10*u.au), ('e', 1.5), ('i', -10*u.deg)]
    for k, v in bad_kws:
        kw = kw1.copy()
        kw[k] = v
        with pytest.raises(ValueError):
            KeplerElements(**kw)

    with pytest.raises(ValueError):
        elems = KeplerElements(P=10*u.day, e=0.5, K=10*u.km/u.s, a=0.1*u.au,
                               omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)

    # Check warning
    with catch_warnings() as w:
        KeplerElements(P=1*u.day, a=5*u.au, e=0.,
                       omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)
    assert len(w) > 0


def test_twobodykepler():

    TwoBodyKeplerElements(P=10*u.day, e=0.5,
                          m1=1*u.Msun, m2=5*u.Msun,
                          omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)
    TwoBodyKeplerElements(a=10*u.au, e=0.5,
                          m1=1*u.Msun, m2=5*u.Msun,
                          omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)

    # check attributes
    elems = TwoBodyKeplerElements(P=10*u.day, e=0.5,
                                  m1=5*u.Msun, m2=1*u.Msun,
                                  omega=10*u.deg, i=20*u.deg, Omega=30*u.deg)

    with pytest.raises(AttributeError):
        elems.K

    with pytest.raises(AttributeError):
        elems.m_f

    elems.get_body('1')
    elems.get_body('2')

    elems.get_body(1)
    elems.get_body(2)

    with pytest.raises(ValueError):
        elems.get_body('derp')



import astropy.units as u
import numpy as np

from .elements import _parse_time

__all__ = ['RVTrend', 'PolynomialRVTrend']

class RVTrend:
    pass

class PolynomialRVTrend(RVTrend):

    def __init__(self, coeffs=None, t0=None):
        """Specify a polynomial trend in radial velocities by specifying the
        polynomial coefficients and optionally a reference time.

        Note that the coefficients are expected in the reverse order of those
        expected by numpy's ``poly`` functionality, i.e. ``coeffs[0]`` is the
        constant term, ``coeffs[1]`` is the linear term, etc.

        Parameters
        ----------
        coeffs : iterable
            A tuple or list of `~astropy.units.Quantity` objects that have the
            appropriate velocity per time units.
        t0 : `astropy.time.Time`, numeric (optional)
            The reference time for the trend.
        """
        if coeffs is None:
            coeffs = []

        _coeffs = []

        for i, coeff in enumerate(coeffs):
            if not hasattr(coeff, 'unit'):
                raise ValueError("Input coefficients must be a Quantity with "
                                 "velocity / time^(-index) units!")

            if i == 0:
                self._v_unit = coeff.unit # record the unit of the constant term
            _unit = self._v_unit / (u.day ** i)

            if not coeff.unit.is_equivalent(_unit):
                raise u.UnitsError("Input coefficients must have velocity / "
                                   "time^(-index) units!")

            _coeffs.append(coeff.to(_unit))

        self.coeffs = tuple(_coeffs)

        if t0 is not None:
            t0 = _parse_time(t0)
        self.t0 = t0

    def __call__(self, t):
        """Evaluate the predicted velocity at the specified time."""

        t = _parse_time(t)

        if len(self.coeffs) == 0:
            return np.zeros_like(t).astype(float)

        if self.t0 is not None:
            t = (t - self.t0).jd
        else:
            t = t.tcb.mjd

        coeffs = []
        for i, x in enumerate(self.coeffs):
            coeffs.append(x.to_value(self._v_unit / u.day**i))

        return np.polyval(coeffs[::-1], t) * self._v_unit



# Third-party
from astropy.time import Time
from astropy.utils import check_broadcast
import numpy as np

__all__ = ['ArrayProcessor']


class ArrayProcessor:

    def __init__(self, *arrs):
        self.arrs = [np.array(arr) for arr in arrs]

    def prepare_arrays(self):
        """Make sure input arrays are all C-contiguous and have same shape."""
        self.max_shape = None
        for arr in self.arrs:
            if self.max_shape is None:
                self.max_shape = arr.shape
            elif arr.shape > self.max_shape:
                self.max_shape = arr.shape

        orig_shapes = []
        arrs_1d = []
        for arr in self.arrs:
            orig_shapes.append(arr.shape)
            arr = np.broadcast_to(arr, self.max_shape).ravel()
            arrs_1d.append(np.ascontiguousarray(arr.astype(np.float64)))

        if not check_broadcast(orig_shapes):
            raise ValueError("Shapes are not broadcastable: {0}"
                             .format(orig_shapes))

        return arrs_1d

    def prepare_result(self, res):
        return res.reshape(self.max_shape)


def format_doc(docstring, *args, **kwargs):
    """
    A modified version of `astropy.utils.decorators.format_doc` that first
    applies ``.format()`` with ``__doc__``, then calls ``.format()`` with the
    arguments.

    Replaces the docstring of the decorated object and then formats it.
    The formatting works like :meth:`str.format` and if the decorated object
    already has a docstring this docstring can be included in the new
    documentation if you use the ``{__doc__}`` placeholder.
    Its primary use is for reusing a *long* docstring in multiple functions
    when it is the same or only slightly different between them.
    Parameters
    ----------
    docstring : str or object or None
        The docstring that will replace the docstring of the decorated
        object. If it is an object like a function or class it will
        take the docstring of this object. If it is a string it will use the
        string itself. One special case is if the string is ``None`` then
        it will use the decorated functions docstring and formats it.
    args :
        passed to :meth:`str.format`.
    kwargs :
        passed to :meth:`str.format`. If the function has a (not empty)
        docstring the original docstring is added to the kwargs with the
        keyword ``'__doc__'``.
    Raises
    ------
    ValueError
        If the ``docstring`` (or interpreted docstring if it was ``None``
        or not a string) is empty.
    IndexError, KeyError
        If a placeholder in the (interpreted) ``docstring`` was not filled. see
        :meth:`str.format` for more information.
    Notes
    -----
    Using this decorator allows, for example Sphinx, to parse the
    correct docstring.
    """
    def set_docstring(obj):
        if docstring is None:
            # None means: use the objects __doc__
            doc = obj.__doc__
            # Delete documentation in this case so we don't end up with
            # awkwardly self-inserted docs.
            obj.__doc__ = None
        elif isinstance(docstring, str):
            # String: use the string that was given
            doc = docstring
        else:
            # Something else: Use the __doc__ of this
            doc = docstring.__doc__

        if not doc:
            # In case the docstring is empty it's probably not what was wanted.
            raise ValueError('docstring must be a string or containing a '
                             'docstring that is not empty.')

        # If the original has a not-empty docstring append it to the format
        # kwargs.
        _doc = obj.__doc__ or ''
        doc = doc.format(__doc__=_doc)
        obj.__doc__ = doc.format(*args, **kwargs)
        return obj
    return set_docstring



# Third-party
import astropy.units as u
import astropy.coordinates as coord
from astropy.coordinates import (frame_transform_graph, CoordinateAttribute,
                                 FunctionTransform, AffineTransform, ICRS)
from astropy.coordinates.matrix_utilities import (rotation_matrix,
                                                  matrix_product,
                                                  matrix_transpose)

__all__ = ['ReferencePlaneFrame']

_cache = {}


def _make_cls(framecls):
    """Analogous to
    astropy.coordinates.builtin_frames.skyoffset._make_skyoffset_cls -- see
    notes in there.
    """

    if framecls in _cache:
        return _cache[framecls]

    # the class of a class object is the metaclass
    framemeta = framecls.__class__

    class ReferencePlaneMeta(framemeta):
        """
        This metaclass renames the class to be "ReferencePlane<framecls>" and
        also adjusts the frame specific representation info so that spherical
        names are always "lon" and "lat" (instead of, e.g., "ra" and "dec").
        """

        def __new__(cls, name, bases, members):
            # Only 'origin' is needed here, to set the origin frame properly.
            members['origin'] = CoordinateAttribute(frame=framecls,
                                                    default=None)

            # This has to be done because FrameMeta will set these attributes
            # to the defaults from BaseCoordinateFrame when it creates the base
            # SkyOffsetFrame class initially.
            members['_default_representation'] = framecls._default_representation
            members['_default_differential'] = framecls._default_differential

            newname = name[:-5] if name.endswith('Frame') else name
            newname += framecls.__name__

            return super().__new__(cls, newname, bases, members)

    _ReferencePlaneFramecls = ReferencePlaneMeta('ReferencePlaneFrame',
        (ReferencePlaneFrame, framecls),
        {'__doc__': ReferencePlaneFrame.__doc__})

    @frame_transform_graph.transform(FunctionTransform,
                                     _ReferencePlaneFramecls,
                                     _ReferencePlaneFramecls)
    def referenceplane_to_referenceplane(from_coord, to_frame):
        """Transform between two reference plane frames."""

        if not from_coord.origin.has_data or not to_frame.origin.has_data:
            raise ValueError("Reference plane origin frame object must have "
                             "data, i.e. it must have a position specified.")

        # This transform goes through the parent frames on each side.
        # from_frame -> from_frame.origin -> to_frame.origin -> to_frame
        intermediate_from = from_coord.transform_to(from_coord.origin)
        intermediate_to = intermediate_from.transform_to(to_frame.origin)
        return intermediate_to.transform_to(to_frame)

    @frame_transform_graph.transform(AffineTransform, _ReferencePlaneFramecls,
                                     framecls)
    def referenceplane_to_coord(reference_plane_coord, to_frame):
        """Convert a reference plane coordinate to the original system"""

        if (reference_plane_coord.origin is None or
                not reference_plane_coord.origin.has_data):
            raise ValueError("Reference plane origin frame object must have "
                             "data, i.e. it must have a position specified.")

        origin = reference_plane_coord.origin.spherical
        if origin.distance.unit == u.dimensionless_unscaled:
            dist = 0. * u.pc
        else:
            dist = origin.distance
        sun = coord.CartesianRepresentation([0, 0, 1] * dist)

        mat1 = rotation_matrix(90*u.deg+origin.lat, 'y')
        mat2 = rotation_matrix(-origin.lon, 'z')
        M = matrix_product(mat2, mat1)

        offset = (-sun).transform(M)
        return M, offset

    @frame_transform_graph.transform(AffineTransform, framecls,
                                     _ReferencePlaneFramecls)
    def coord_to_referenceplane(from_coord, to_reference_plane):
        """Convert a sky coordinate to a reference-plane frame."""

        if (to_reference_plane.origin is None or
                not to_reference_plane.origin.has_data):
            raise ValueError("Reference plane origin frame object must have "
                             "data, i.e. it must have a position specified.")

        M, offset = referenceplane_to_coord(to_reference_plane, from_coord)
        M = matrix_transpose(M)
        offset = (-offset).transform(M)
        return M, offset

    _cache[framecls] = _ReferencePlaneFramecls
    return _ReferencePlaneFramecls


class ReferencePlaneFrame(coord.BaseCoordinateFrame):
    """A coordinate frame aligned with the reference plane coordinates of a
    Kepler orbit, centered on the barycenter or reference point of the orbit.

    See :ref:`celestial-reference-plane` for more information.

    ``ReferencePlaneFrame`` objects always have generic component names for
    spherical coordinates of ``lon``/``lat``, *not* the component names for the
    frame of the ``origin``.

    Parameters
    ----------
    representation : `BaseRepresentation` or None
        A representation object or None to have no data (or use the other
        keywords).
    origin : `SkyCoord` or low-level coordinate object.
        The coordinate which specifies the origin of this frame. This is
        typically a sky position and a distance to the barycenter or reference
        point of the orbit at a particular epoch.

    Notes
    -----
    ``ReferencePlaneFrame`` is a factory class.  That is, the objects that it
    yields are *not* actually objects of class ``ReferencePlaneFrame``.
    Instead, distinct classes are created on-the-fly for whatever the frame
    class is of ``origin``.
    """

    origin = CoordinateAttribute(default=None, frame=None)

    def __new__(cls, *args, **kwargs):
        # We don't want to call this method if we've already set up
        # an skyoffset frame for this class.
        if not (issubclass(cls, ReferencePlaneFrame) and
                cls is not ReferencePlaneFrame):
            # We get the origin argument, and handle it here. Default is ICRS:
            # the user might want to use arbitrary reference plane coordinates
            # without every transforming them
            origin_frame = kwargs.get('origin', ICRS())

            if hasattr(origin_frame, 'frame'):
                origin_frame = origin_frame.frame
            newcls = _make_cls(origin_frame.__class__)
            return newcls.__new__(newcls, *args, **kwargs)

        # http://stackoverflow.com/questions/19277399/why-does-object-new-work-differently-in-these-three-cases
        # See above for why this is necessary. Basically, because some child
        # may override __new__, we must override it here to never pass
        # arguments to the object.__new__ method.
        if super().__new__ is object.__new__:
            return super().__new__(cls)
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.origin is not None and not self.origin.has_data:
            raise ValueError('The origin supplied to ReferencePlaneFrame has '
                             'no data.')
        if self.has_data and hasattr(self.data, 'lon'):
            self.data.lon.wrap_angle = 180*u.deg



from collections import defaultdict
from distutils.core import Extension


def get_extensions():
    import numpy as np

    exts = []

    # malloc
    mac_incl_path = "/usr/include/malloc"

    cfg = defaultdict(list)
    cfg['include_dirs'].append(np.get_include())
    cfg['include_dirs'].append(mac_incl_path)
    cfg['include_dirs'].append('twobody/')
    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('twobody/wrap.pyx')
    cfg['sources'].append('twobody/src/twobody.c')
    exts.append(Extension('twobody.wrap', **cfg))

    return exts




# Third-party
from astropy.time import Time
import astropy.coordinates as coord
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import numpy as np
import pytest

try:
    import matplotlib.pyplot as plt
    HAS_MPL = True
except ImportError:
    HAS_MPL = False

# Package
from ..orbit import KeplerOrbit
from ..barycenter import Barycenter
from ..bary_trends import PolynomialRVTrend

def get_random_orbit_pars(rnd=None, barycenter=None):
    if rnd is None:
        rnd = np.random.RandomState()

    P = 10**rnd.uniform(0, 3)*u.day
    K = np.abs(rnd.normal(0, 100))*u.km/u.s
    e = rnd.uniform(0, 1)
    i = rnd.uniform(0, 180)*u.deg
    n = 2*np.pi/P
    a = K * np.sqrt(1-e**2) / (n * np.sin(i))
    return dict(P=P, a=a, e=e, i=i,
                omega=rnd.uniform(0, 360)*u.deg,
                Omega=rnd.uniform(0, 360)*u.deg,
                M0=rnd.uniform(0, 360)*u.deg,
                barycenter=barycenter)


def test_radial_velocity():
    # Compare the short-circuit RV to the full calculation

    rnd = np.random.RandomState(seed=42)
    n_iter = 32

    bc_kw = dict(ra=210.413231*u.deg,
                 dec=-61.3412*u.deg,
                 distance=101*u.pc)

    # at rest w.r.t. sun
    bc1 = Barycenter(origin=coord.ICRS(**bc_kw,
                                       pm_ra_cosdec=0*u.mas/u.yr,
                                       pm_dec=0.*u.mas/u.yr,
                                       radial_velocity=0*u.km/u.s),
                     t0=Time('J2018.0'))

    bc2 = Barycenter(origin=coord.ICRS(**bc_kw,
                                       pm_ra_cosdec=1000*u.mas/u.yr,
                                       pm_dec=1000.*u.mas/u.yr,
                                       radial_velocity=-260*u.km/u.s),
                     t0=Time('J2018.0'))

    times = Time('J2018.0') + np.linspace(-100, 100, 128) * u.day

    # Even when barycenter is moving, rv1 and rv2 should be the same
    for n in range(n_iter):
        orb = KeplerOrbit(**get_random_orbit_pars(rnd, bc2))

        rv1 = orb.radial_velocity(times)
        _rp = orb.reference_plane(times)
        rv2 = -_rp.cartesian.differentials['s'].d_z + bc2.origin.radial_velocity

        assert quantity_allclose(rv1, rv2, rtol=0, atol=1E-4*u.m/u.s)

    # All RV's should be equivalent
    for n in range(n_iter):
        orb = KeplerOrbit(**get_random_orbit_pars(rnd, bc1))

        rv1 = orb.radial_velocity(times)
        _rp = orb.reference_plane(times)
        rv2 = -_rp.cartesian.differentials['s'].d_z
        rv3 = orb.icrs(times).radial_velocity

        assert quantity_allclose(rv1, rv2, rtol=0, atol=1E-4*u.m/u.s)
        # We expect true differences because of projection effects
        assert quantity_allclose(rv1, rv3, rtol=0, atol=1*u.km/u.s)


def test_velocity_trend():
    rnd = np.random.RandomState(seed=42)

    coeffs = [0*u.km/u.s,
              0*u.km/u.s/u.day,
              1e-2*u.km/u.s/u.day**2]
    trend = PolynomialRVTrend(coeffs,
                              t0=Time('J2018.0'))
    pars = get_random_orbit_pars(rnd, trend)
    pars['a'] = 0*u.au
    orbit = KeplerOrbit(**pars)

    t = Time('J2018.0') + np.linspace(-100, 100, 256)*u.day
    rv = orbit.radial_velocity(t)
    assert quantity_allclose(rv[0], rv[-1])
    assert quantity_allclose(rv[0], 100.*u.km/u.s)

    # OK THIS SUCKS (copy pasta)
    coeffs = [0*u.km/u.s,
              1*u.km/u.s/u.day,
              0*u.km/u.s/u.day**2]
    trend = PolynomialRVTrend(coeffs,
                              t0=Time('J2018.0'))
    pars = get_random_orbit_pars(rnd, trend)
    pars['a'] = 0*u.au
    orbit = KeplerOrbit(**pars)

    t = Time('J2018.0') + np.linspace(-100, 100, 256)*u.day
    rv = orbit.radial_velocity(t)
    assert quantity_allclose(rv[0], -rv[-1])
    assert quantity_allclose(rv[0], -100.*u.km/u.s)


@pytest.mark.skipif(not HAS_MPL, reason="matplotlib not installed")
def test_plotting():
    orbit = KeplerOrbit(P=100*u.day, a=1.*u.au, e=0.1,
                        omega=0*u.deg, i=41*u.deg, Omega=0*u.deg)

    mjd = np.linspace(56823.123, 57293.2345, 1024)  # random MJD's
    t = Time(mjd, format='mjd', scale='utc')

    # plotting
    for _t in [mjd, t]:
        orbit.plot_rv(_t)
        orbit.plot_rv(_t, t_kwargs=dict(format='jd', scale='utc'))

    fig,ax = plt.subplots(1,1)
    orbit.plot_rv(time=t, ax=ax, plot_kwargs=dict(color='#de2d26'))

    fig,ax = plt.subplots(1,1)
    orbit.plot_rv(time=mjd, ax=ax)

    plt.close('all')


def test_nan():

    mjd = np.linspace(56123.123, 57293.2345, 1024)  # random MJD's
    t = Time(mjd, format='mjd', scale='utc')

    orb = KeplerOrbit(P=1.5*u.year, e=0.67, M0=0*u.deg, omega=17.14*u.deg,
                      i=np.nan*u.deg, Omega=np.nan*u.deg, a=np.nan*u.au)
    rv = orb.radial_velocity(t)
    urv = orb.unscaled_radial_velocity(t)

    assert np.all(np.isnan(rv.value))
    assert np.all(np.isfinite(urv))

    orb = KeplerOrbit(P=1.5*u.year, e=0.67, M0=0*u.deg, omega=17.14*u.deg)
    with pytest.raises(ValueError):
        rv = orb.radial_velocity(t)
    urv = orb.unscaled_radial_velocity(t)

    assert np.all(np.isnan(rv.value))
    assert np.all(np.isfinite(urv))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
A package for computing orbits and astronomical observables for binary stars,
exoplanets, and other gravitational two-body systems.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

# Enforce Python version check during package import.
# This is the same check as the one at the top of setup.py
import sys

__minimum_python_version__ = "3.6"

class UnsupportedPythonError(Exception):
    pass

if sys.version_info < tuple((int(val) for val in __minimum_python_version__.split('.'))):
    raise UnsupportedPythonError("packagename does not support Python < {}"
                                 .format(__minimum_python_version__))

if not _ASTROPY_SETUP_:
    from .anomaly import *
    from .barycenter import *
    from .bary_trends import *
    from .elements import *
    from .orbit import *
    from .reference_plane import *
    from .transforms import *
    from .units import *



# Standard library
import abc
import warnings

# Third-party
import astropy.coordinates as coord
from astropy.constants import G, c
from astropy.time import Time
import astropy.units as u
import numpy as np
from numpy import pi

# Project
from .transforms import a_m_to_P, P_m_to_a, PeKi_to_a
from .units import UnitSystem

__all__ = ['OrbitalElements', 'KeplerElements', 'TwoBodyKeplerElements']


def _parse_time(t):
    if not isinstance(t, Time):
        return Time(t, format='mjd', scale='tcb')
    return t


class ElementsMeta(abc.ABCMeta):

    def __new__(mcls, name, bases, members):

        if 'names' not in members:
            raise ValueError('OrbitalElements subclasses must contain a '
                             'defined class attribute "names" that specified '
                             'the string names of the elements.')

        for name_ in members['names']:
            mcls.readonly_prop_factory(members, name_)

        return super().__new__(mcls, name, bases, members)

    @staticmethod
    def readonly_prop_factory(members, attr_name):
        def getter(self):
            return self.units.decompose(getattr(self, '_' + attr_name))
        members[attr_name] = property(getter)


class OrbitalElements(metaclass=ElementsMeta):
    """
    Subclasses must define the class attribute ``.default_units`` to be a
    ``UnitSystem`` instance.
    """

    names = []

    def __init__(self, units):

        # Make sure the units specified are a UnitSystem instance
        if units is None:
            units = self.default_units

        elif units is not None and not isinstance(units, UnitSystem):
            units = UnitSystem(*units)

        self.units = units

        # Now make sure all element name attributes have been set:
        for name in self.names:
            if not hasattr(self, '_' + name):
                raise AttributeError('Invalid class definition!')


class BaseKeplerElements(OrbitalElements):
    default_units = UnitSystem(u.au, u.day, u.Msun, u.degree, u.km / u.s)
    names = ['P', 'a', 'e', 'omega', 'i', 'Omega', 'M0']

    def __init__(self, P=None, a=None,
                 e=0, omega=None, i=None, Omega=None,
                 M0=None, t0=None, units=None):
        """Base class for Keplerian orbital elements.

        Parameters
        ----------
        P : quantity_like [time]
            Orbital period.
        a : quantity_like [length] (optional)
            Semi-major axis. If unspecified, computed orbits will be unscaled.
        e : numeric (optional)
            Orbital eccentricity. Default is circular, ``e=0``.
        omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Argument of pericenter.
        i : quantity_like, `~astropy.coordinates.Angle` [angle]
            Inclination of the orbit.
        Omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Longitude of the ascending node.
        M0 : quantity_like, `~astropy.coordinates.Angle` [angle] (optional)
            Mean anomaly at epoch ``t0``. Default is 0º if not specified.
        t0 : numeric, `~astropy.coordinates.Time` (optional)
            Reference epoch. If a number is passed in, it is assumed to be
            a solar system barycentric modified julian date (BMJD). The default
            is J2000 if not specified.
        units : `~twobody.units.UnitSystem`, iterable (optional)
            The unit system to represent quantities in. The default unit system
            is accessible as `KeplerElements.default_units`.

        """

        if M0 is None:
            # Default phase at reference epoch is 0º
            M0 = 0 * u.degree

        if t0 is None:
            # Default reference epoch is J2000
            t0 = Time('J2000')

        t0 = _parse_time(t0)

        # Now check that required elements are defined:
        _required = ['P', 'omega']
        for name in _required:
            if eval(name) is None:
                raise ValueError("You must specify {0}.".format(name))

        # Value validation:
        if P < 0 * u.day:
            raise ValueError("Period `P` must be positive.")

        if a is not None and a < 0 * u.au:
            raise ValueError("Semi-major axis `a` must be positive.")

        if e < 0 or e >= 1:
            raise ValueError("Eccentricity `e` must be: 0 <= e < 1")

        if i is not None and (i < 0*u.deg or i > 180 * u.deg):
            raise ValueError("Inclination `i` must be between 0º and 180º, you "
                             "passed in i={:.3f}".format(i.to(u.degree)))

        if i is None:
            i = np.nan * u.deg

        if Omega is None:
            Omega = np.nan * u.deg

        # Set object attributes, but make them read-only
        self._a = a if a is not None else 1. * u.dimensionless_unscaled
        self._P = P
        self._e = float(e) * u.dimensionless_unscaled
        self._omega = coord.Angle(omega).wrap_at(360 * u.deg)
        self._i = coord.Angle(i)
        self._Omega = coord.Angle(Omega).wrap_at(360 * u.deg)
        self._M0 = coord.Angle(M0)
        self.t0 = t0

        # Must happen at the end because it validates that all element names
        # have been set properly:
        super().__init__(units=units)


class KeplerElements(BaseKeplerElements):
    names = ['P', 'a', 'e', 'omega', 'i', 'Omega', 'M0']

    @u.quantity_input(P=u.year, a=u.au, K=u.km/u.s,
                      omega=u.deg, i=u.deg, Omega=u.deg, M0=u.deg)
    def __init__(self, *, P=None, a=None, K=None,
                 e=0, omega=None, i=None, Omega=None,
                 M0=None, t0=None, units=None):
        """Keplerian orbital elements for a single orbit.

        The elements are assumed to be relative to an inertial frame, typically
        the barycenter of a two-body system.

        Parameters
        ----------
        P : quantity_like [time]
            Orbital period.
        a : quantity_like [length] (optional)
            Semi-major axis. Specify this OR the semi-amplitude ``K``, but not
            both. If unspecified, computed orbits will be unscaled.
        K : quantity_like [speed] (optional)
            Velocity semi-amplitudes. Specify this OR the semi-major axis ``a``,
            but not both. If unspecified, computed orbits will be unscaled.
        e : numeric (optional)
            Orbital eccentricity. Default is circular, ``e=0``.
        omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Argument of pericenter.
        i : quantity_like, `~astropy.coordinates.Angle` [angle]
            Inclination of the orbit.
        Omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Longitude of the ascending node.
        M0 : quantity_like, `~astropy.coordinates.Angle` [angle] (optional)
            Mean anomaly at epoch ``t0``. Default is 0º if not specified.
        t0 : numeric, `~astropy.coordinates.Time` (optional)
            Reference epoch. If a number is passed in, it is assumed to be
            a solar system barycentric modified julian date (BMJD). The default
            is J2000 if not specified.
        units : `~twobody.units.UnitSystem`, iterable (optional)
            The unit system to represent quantities in. The default unit system
            is accessible as `KeplerElements.default_units`.

        """
        if K is not None and a is not None:
            raise ValueError("Both `K` and `a` were specified, but can only "
                             "accept one or the other.")

        super().__init__(P=P, a=a, e=e, omega=omega, i=i, Omega=Omega,
                         M0=M0, t0=t0, units=units)

        # TODO: this is a bit messy because we double-initialize! The first
        # time to validate, the second to use the proper "a"
        if K is not None:
            a = PeKi_to_a(P, e, K, i)
            super().__init__(P=P, a=a, e=e, omega=omega, i=i, Omega=Omega,
                             M0=M0, t0=t0, units=units)

        if self.K.unit.physical_type == 'speed':
            # Now we do a quick validation to make sure we're not relativistic...
            _K_c = (self.K / c).decompose()
            if _K_c > 1E-2:
                warnings.warn('Velocity semiamplitude is large enough that '
                              'relativistic effects are important: K/c = {:.2f}'
                              ' but are not currently accounted for in TwoBody'
                              .format(_K_c), RuntimeWarning)

    @property
    def K(self):
        """Velocity semi-amplitude."""
        K = 2*pi * self.a * np.sin(self.i) / (self.P * np.sqrt(1 - self.e**2))
        return self.units.decompose(K)

    @property
    def m_f(self):
        """Binary mass function."""
        mf_circ = self.P * self.K**3 / (2*pi * G)
        return self.units.decompose(mf_circ) * (1 - self.e**2)**1.5

    # Python builtins
    def __repr__(self):
        return ("<KeplerElements [P={:.2f}, a={:.2f}, e={:.2f}, "
                "ω={:.2f}, i={:.2f}, Ω={:.2f}]>"
                .format(self.P, self.a, self.e, self.omega, self.i, self.Omega))


class TwoBodyKeplerElements(BaseKeplerElements):
    names = ['P', 'a', 'e', 'm1', 'm2', 'omega', 'i', 'Omega', 'M0']

    @u.quantity_input(P=u.year, a=u.au, m1=u.Msun, m2=u.Msun,
                      omega=u.deg, i=u.deg, Omega=u.deg, M0=u.deg)
    def __init__(self, *, P=None, a=None, m1=None, m2=None,
                 e=0, omega=None, i=None, Omega=None,
                 M0=None, t0=None, units=None):
        """Keplerian orbital elements for a two-body system.

        You can either specify period, ``P`` and the masses ``m1``, ``m2``, or
        the separation ``a`` and the two masses.

        Parameters
        ----------
        P : quantity_like [time]
            Orbital period.
        a : quantity_like [length]
            Semi-major axis of the fictitious particle
        e : numeric (optional)
            Orbital eccentricity. Default is circular, ``e=0``.
        omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Argument of pericenter.
        i : quantity_like, `~astropy.coordinates.Angle` [angle]
            Inclination of the orbit.
        Omega : quantity_like, `~astropy.coordinates.Angle` [angle]
            Longitude of the ascending node of the fictitious particle.
        M0 : quantity_like, `~astropy.coordinates.Angle` [angle] (optional)
            Mean anomaly at epoch ``t0``. Default is 0º if not specified.
        t0 : numeric, `~astropy.coordinates.Time` (optional)
            Reference epoch. Default is J2000 if not specified.
        units : `~twobody.units.UnitSystem`, iterable (optional)
            The unit system to represent quantities in. The default unit system
            is accessible as `KeplerElements.default_units`.

        """

        if m1 is None or m2 is None:
            raise ValueError("You must specify m1 and m2.")

        if P is not None and a is not None:
            raise ValueError("You can only specify one of period `P` or "
                             "semi-major axis `a`.")

        if P is None:
            P = a_m_to_P(a, m1 + m2)

        if a is None:
            a = P_m_to_a(P, m1 + m2)

        self._m1 = m1
        self._m2 = m2
        self.m_tot = m1 + m2

        # values are validated
        super().__init__(a=a, P=P,
                         e=e, omega=omega, i=i, Omega=omega,
                         M0=M0, t0=t0, units=units)

    def get_body(self, num):
        """Get the orbital elements for the specified body.

        Parameters
        ----------
        num : str ('1' or '2')
            Get the orbital elements of the primary `'1'` or secondary `'2'`.

        Returns
        -------
        elements : `twobody.KeplerElements`

        """
        num = str(num)

        if num == '1':
            a = self.m2 / self.m_tot * self.a
            omega = self.omega

        elif num == '2':
            a = self.m1 / self.m_tot * self.a
            omega = self.omega + np.pi*u.radian

        else:
            raise ValueError("Invalid input '{0}' - must be '1' or '2'"
                             .format(num))

        return KeplerElements(a=a, P=self.P,
                              e=self.e, omega=omega, i=self.i, Omega=self.Omega,
                              M0=self.M0, t0=self.t0)

    @property
    def primary(self):
        """Shorthand for ``.get_body('1')``. Returns the orbital elements for
        the primary body, which corresponds to ``m1`` (not the more massive of
        the two).
        """
        return self.get_body('1')

    @property
    def secondary(self):
        """Shorthand for ``.get_body('2')``. Returns the orbital elements for
        the secondary body, which corresponds to ``m2`` (not the less massive of
        the two).
        """
        return self.get_body('2')

    # Python builtins
    def __repr__(self):
        return ("<TwoBodyKeplerElements [m1={:.2f}, m2={:.2f}, "
                "P={:.2f}, a={:.2f}, e={:.2f}, "
                "ω={:.2f}, i={:.2f}, Ω={:.2f}]>"
                .format(self.m1, self.m2, self.P, self.a, self.e,
                        self.omega, self.i, self.Omega))



# Third-party
import astropy.units as u
from astropy.utils.misc import isiterable
from astropy.constants import G
import numpy as np

# Project
from .utils import format_doc


__all__ = ['a_P_to_m', 'a_m_to_P', 'P_m_to_a', 'get_m2_min']


doc_a = """a : quantity_like [length]
        Semi-major axis.
"""

doc_P = """P : quantity_like [time]
        Orbital period.
"""

doc_m = """m : quantity_like [mass]
        Total mass.
"""


@u.quantity_input(a=u.au, P=u.day)
@format_doc("""{__doc__}""", a=doc_a, P=doc_P)
def a_P_to_m(a, P):
    """Compute the total mass given the semi-major axis and period.

    Parameters
    ----------
    {a}
    {P}
    """
    return a**3 / G * (2*np.pi/P)**2


@u.quantity_input(a=u.au, m=u.Msun)
@format_doc("""{__doc__}""", a=doc_a, m=doc_m)
def a_m_to_P(a, m):
    """Compute the orbital period given the semi-major axis and total mass.

    Parameters
    ----------
    {a}
    {m}
    """
    return 2*np.pi * np.sqrt(a**3 / (G * m))


@u.quantity_input(P=u.day, m=u.Msun)
@format_doc("""{__doc__}""", P=doc_P, m=doc_m)
def P_m_to_a(P, m):
    """Compute the semi-major axis given the orbital period and total mass.

    Parameters
    ----------
    {P}
    {m}
    """
    return np.cbrt(G * m * (P/(2*np.pi))**2)


@u.quantity_input(P=u.day, K=u.km/u.s, i=u.deg)
@format_doc("""{__doc__}""", P=doc_P)
def PeKi_to_a(P, e, K, i=None):
    """Compute the semi-major axis given the orbital period, eccentricity,
    semi-amplitude, and inclination. If you don't have a measured inclination,
    use the default i=90º and the returned semi-major axis will be a*sin(i)
    instead.

    Parameters
    ----------
    {P}
    e : numeric
        Eccentricity.
    K : quantity_like [speed]
        Velocity semi-amplitude.
    i : quantity_like [angle]
        Inclination.

    Returns
    -------
    a : `~astropy.units.Quantity` [length]
        Semi-major axis.
    """
    if i is None:
        i = 90*u.deg
    a = K / (2*np.pi * np.sin(i)) * (P * np.sqrt(1 - e**2))
    return a


def _m2_func(m2, m1, sini, mf):
    return (m2*sini)**3 / (m1 + m2)**2 - mf


@u.quantity_input(m1=u.Msun, mf=u.Msun)
def get_m2_min(m1, mf):
    """Compute the minimum companion mass given the primary mass and the binary
    mass function.

    Parameters
    ----------
    m1 : quantity_like [mass]
        Primary mass.
    mf : quantity_like [mass]
        Binary mass function.

    Returns
    -------
    m2_min : `~astropy.units.Quantity` [mass]
        The minimum companion mass.
    """
    from scipy.optimize import root

    mf = mf.to(m1.unit)
    if isiterable(m1) and isiterable(mf):
        m2s = []
        for x, y in zip(m1, mf):
            try:
                res = root(_m2_func, x0=10., args=(x.value, 1., y.value))
                if not res.success:
                    raise RuntimeError('Unsuccessful')
                m2s.append(res.x[0])
            except Exception as e:
                m2s.append(np.nan)
        return m2s * m1.unit

    else:
        res = root(_m2_func, x0=10., args=(m1.value, 1., mf.value))

        if res.success:
            return res.x[0] * m1.unit
        else:
            return np.nan * m1.unit
